/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Kyle Nitzsche <kyle.nitzsche@canonical.com>
 *              Scott Sweeny <scott.sweeny@canonical.com>
 */

#ifndef CHILDSCOPE_H
#define CHILDSCOPE_H

#include <functional>

#include <unity-scopes.h>
#include <unity/scopes/ScopeMetadata.h>
#include <unity/scopes/RegistryProxyFwd.h>

#include <QDebug>
#include <QString>

namespace us = unity::scopes;


/**
    \brief The child scope to be aggregated
    Provides methods to enable common child scope customizations
*/
class AggChildScope
{
public:
    /**
     \brief Constructor
     \param id the fully qualified scope id (for example: com.canonical.scopes.photos-local_photos_local)
     \param query_string the Query's query().query_string()
     \param use_category_link_to_child_ Optional. When false the category containing the child scope cannot be tapped to open the child scope

    */
    AggChildScope(const std::string id, const std::string query_string, bool use_category_link_to_child_ = true);

    /**
     \brief The displayed category title
    \param title_ The title
    */
    void set_category_title(const std::string title_);

    /**
    \brief Return the Category title
    \return title
    */
    std::string category_title();

    /**
    \brief Set Category
    \return Category
    */
    void set_category(us::Category::SCPtr category_);

    /**
    \brief Return SCPtr to Category
    \return Category
    */
    us::Category::SCPtr category();

    /**
    \brief The json defining the CategoryRenderer template used when surfacing results.
    \param surface_template The json passed to the CategoryRenderer
    */
    void set_surface_template(const std::string surface_template);

    /**
    \brief Return the json defining the CategoryRenderer template used when surfacing results.
    \param surface_template Return the json of the surfacing CategoryRenderer
    */
    std::string surface_template();

    /**
    \brief Return the json defining the CategoryRenderer template used when displaying search results.
    \param search_template Return the json of the search CategoryRenderer
    */
    std::string search_template();

    /**
    \brief The json defining the CategoryRenderer template used when displaying search results.
    \param search_template The json passed to the CategoryRenderer
    */
    void set_search_template(const std::string search_template);

    /**
    \brief Display child scope results only from the specific department
    \param department_id The ID of the child scope department.
    */
    void set_child_department(const std::string department_id_);

    /**
    \brief Return child department id
    \return Return child department id
    */
    std::string child_department();

    /**
    \brief Display child scope results only from the specified category
    \parm category_id The ID of the cagtegory of the result.
    */
    void set_child_category(const std::string category_id);

    /**
    \brief Return the category_id. Undefined it not using set_child_category()
    \return The child_category
    */
    std::string child_category();
    /**
    \brief Limit child scope results to this number.
    Note: Only use this when also using set_child_category(), otherwise use set_cardinality()
    \parm max The maximum number of results to display.
    */
    void set_category_max_results(int max);

    /**
    \brief Return whether scope is using category_max_results.
    \return Whether scope is using category_max_results.
    */
    bool using_category_max_results();

    /**
    \brief Return the category_max_results.
    \return Return category_max_results.
    */
    int category_max_results();

    /**
    \brief Limit child scope results to this number through scope cardinality.
    Note: Only use this when not using set_child_category(), otherwise use set_category_max_results()
    \parm cardinality The maximum number of results to display.
    */
    void set_cardinality(int cardinality);

    /**
    \brief Return scope cardinality
    \return cardinality
    */
    int cardinality();

    /**
    \brief Return scope ID
    \return The scope ID.
    */
    std::string id();

    /**
    \brief Return True if scope is using categoryLinkToChild, else false 
    \return The scope ID.
    */
    bool category_link_to_child();

    /**
    \brief Return the CannedQuery created when category_link_to_child is set. Undefined if category_link-to-child is not set.
    \return The CannedQuery.
    */
    std::shared_ptr<us::CannedQuery> canned_query();

    /**
    \brief Return True if the AggChildScope scope is installed and registered by id().
    \return True when scope is present.
    */
    bool exists()
    {
        return exists_;
    }

    /**
    \brief Return this scope by ScopeProxy
    */
    us::ScopeProxy proxy();

    /**
    \brief This is called automatically. You generally do not need to call this. If the scope is in the registry, its proxy is saved and its 'exists' bool is made true
    \param A RegistryProxy
    */
    void set_proxy(us::RegistryProxy const);

    /**
    \brief This is called automatically and you generally do not need to call this. Increments the results_idx;
    */
    void inc_result_idx();

    /**
    \brief This is called automatically and you generally do not need to call this. ;
    \return The result_idx`
    */
    int result_idx();

private:

    std::string id_;
    std::string cat_title = "";
    bool using_category_link_to_child;
    std::string dept_id = "";
    bool using_category = false;
    bool using_category_max_results_ = false;
    int category_max_results_;
    int result_idx_ = 0;
    std::string category_id_ = "";
    std::string query_string;
    int cardinality_ = 5;
    us::ScopeMetadata *metadata;
    us::ScopeProxy proxy_;
    bool exists_ = false;
    std::shared_ptr<us::CannedQuery> canned_query_;
    us::Category::SCPtr cat;
    std::string TEMPLATE_NOT_LOGGED_IN;
    std::string TEMPLATE_SURFACE;
    std::string TEMPLATE_SEARCH;
};

#endif
