/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Kyle Nitzsche <kyle.nitzsche@canonical.com>
 *              Scott Sweeny <scott.sweeny@canonical.com>
 *
 */

#include "query.h"
#include "aggchildscope.h"
#include "listener.h"
#include "scope-ids.h"
#include "i18n.h"

#include <locale.h>

#include <unity/scopes/Annotation.h>
#include <unity/scopes/CategorisedResult.h>
#include <unity/scopes/CategoryRenderer.h>
#include <unity/scopes/QueryBase.h>
#include <unity/scopes/SearchMetadata.h>
#include <unity/scopes/SearchReply.h>
#include <unity/scopes/Scope.h>
#include <unity/UnityExceptions.h>

#include <QDir>
#include <QDebug>

QString CALL_ICON;

using namespace unity::scopes;
using namespace std;

QString qstr(std::string);
std::string sstr(QString);

Query::Query(CannedQuery const& query,
                               SearchMetadata const& metadata,
                               const std::string scope_dir,
                               const std::string cache_dir,
                               RegistryProxy const& registry
                               ) :
    SearchQueryBase(query, metadata),
    scope_dir_(scope_dir),
    cache_dir_(cache_dir),
    registry_(registry)
{
    setlocale(LC_ALL, "");
    CALL_ICON = QString("%1/images/placed.png").arg(QString::fromStdString(scope_dir));
}

Query::~Query()
{
}

void Query::cancelled()
{
}

void Query::run(unity::scopes::SearchReplyProxy const& upstream_reply)
{

    if (!registry_) {
        throw ConfigException("No registry available, cannot locate child scopes");
        return;
    }

    std::string query_string = query().query_string();
    const bool empty_search = query_string.empty();

    // if the hints scope has hard coded support for this agg scope, uncomment the following
    /*
    if (check_hints())
    {
        query_string = "";
    }
    */

    // create child scope instances
    make_child_scopes(query_string);

    // for settings-enabled child scopes, complete initializations
    handle_enabled_child_scopes(empty_search, upstream_reply);

    // set up handling for results and handle them, including buffering to ensure category order
    handle_child_scope_results(query_string, empty_search, upstream_reply);

    // if the hints scope has hard coded support for this agg scope, uncomment the following
    /*
    if (handle_hints(upstream_reply))
    {
        return;
    }
    */

}

/**
 * Cycle through scopes in order. For each, set up buffering and handle the current CategorisedResult (or use default handling). Results from the first scope are not buffered, that is they are displayed as soon as they arrive. Results from other scopes are held (buffered) until all preceding scopes have results, thus ensuring scope ordering
 */
void Query::handle_child_scope_results(std::string query_string, bool empty_search, us::SearchReplyProxy const& upstream_reply)
{

    std::vector<std::shared_ptr<ResultForwarder>> replies; // used for buffering/category ordering

    int numHolidays = 0; // we only want one holiday result

    bool buffer = false;
    int idx = -1;//used in scopes_m_int_id to store scope id
    for (const std::string &id : scope_order)
    {

        // get a pointer to the current child scope (by id from the map)
        std::shared_ptr<AggChildScope> scope = scopes_m[id];

        // 'exists' is true if the scpoe it is enabled in settings and present on the system
        if (!scope->exists()) {
            qDebug() << QString("==== scope %1 DOES NOT EXIST or is DISABLED").arg(QString::fromStdString(id));
            continue;
        }

        idx++;
        scopes_m_int_id[idx] = id; // add id to map used later to get dept_id
/*
        bool buffer;
        if (first)
        {
            buffer = false;
        } else
            buffer = true;

         if (replies.size()==0)
            buffer = false;
*/
        qDebug() << QString("==== ADDING SCOPE: %1").arg(QString::fromStdString(id));

        // Add the scope (by its proxy) to our list of scopes
        scopes.push_back(scope->proxy());

        //result handling and buffering for category ordering
        bool use_result = true;

        //if all scopes prior to this scope 


        auto reply = make_shared<Listener>(upstream_reply,buffer,
            [scope, &use_result, this, &upstream_reply](CategorisedResult& res) -> bool {
                scope->inc_result_idx();
                if (scope->using_category_max_results()
                    && scope->result_idx() > scope->category_max_results())
                {
                    use_result = false;
                    return false;
                }
                //TODO: support using child_category AND using a different category for not_logged_in results
                qDebug() << "==== in CAT scope child cat: " << qstr(scope->child_category());
                qDebug() << "==== in CAT res cat: " << qstr(res.category()->id());
                if (scope->child_category() !="" && !qstr(res.category()->id()).startsWith(qstr(scope->child_category()))) {
                    qDebug() << "==== dont use this one: res cat: " << qstr(res.category()->id());
                    use_result = false;
                    return false;
                }
                qDebug() << "==== do use this one: res cat: " << qstr(res.category()->id());
                for (std::shared_ptr<std::pair<std::string,std::string>> swap : this->child_scopes_m[scope->id()]->swap_result_attributes)
                {
                    try
                    {
                        res[swap->second] = res[swap->first];
                    }
                    catch (unity::LogicException)
                    {
                        qDebug() << QString("==== result does not contain %1").arg(qstr(swap->second));
                    }
                }


                // if the result category id is declared as a common_template, use common template 
                if (common_templates.find(child_scopes_m[scope->id()]->result_category_id_to_common_template[res.category()->id()]) != common_templates.end())
                {
                    std::string common_template_id = child_scopes_m[scope->id()]->result_category_id_to_common_template[res.category()->id()];
                    std::string cat_id = common_template_id + scope->id();
                    //qDebug() << "==== common_template_id " << qstr(common_template_id);

                    // if the category has not yet been registered, register it
                    if (registered_common_categories.find(cat_id) == registered_common_categories.end())
                    {
                        std::string template_to_use = common_templates[child_scopes_m[scope->id()]->result_category_id_to_common_template[res.category()->id()]];
                        //qDebug() << "==== template to use:\n" << qstr(template_to_use);
                        registered_common_categories[cat_id] = upstream_reply->register_category(cat_id, scope->category_title(), "",
                                                       CategoryRenderer(template_to_use));
                    }
                    //qDebug() << "==== registered_common_categories: " <<
                    //        qstr(child_scopes_m[scope->id()]->result_category_id_to_common_template[res.category()->id()]);
                    res.set_category(registered_common_categories[cat_id]);
                    //res.set_category(registered_common_categories[res.category()->id()]);
                }
                else //use the child scope specific category
                    res.set_category(scope->category());

                return true;
            });
            if (use_result)
            {
                buffer = true;
                replies.push_back(reply);
            }
//        }

    }

    // create and chain result forwarders to enforce proper order of categores
    for (unsigned int i = 1; i < scopes.size(); ++i) {
        for (unsigned int j = 0; j < i; ++j) {
            replies[j]->add_observer(replies[i]);
        }
    }

    // dispatch search to subscopes
    // also set cardinality, default and per scope
    for (unsigned int i = 0; i < replies.size(); ++i) {
        SearchMetadata metadata(search_metadata());
        if (empty_search && !scopes_m[scopes_m_int_id[i]]->using_category_max_results()) {
            metadata.set_cardinality( scopes_m[scopes_m_int_id[i]]->cardinality() );
        }
        // note by default deptarement_id() is "," which means the top level department/surfaceed content.
        // if targeting a child department (set_child_department(id)),
        // department_id has the appropriate value here
        subsearch(scopes[i], query_string, scopes_m[scopes_m_int_id[i]]->child_department(), FilterState(),
                  metadata, replies[i]);
    }
}
