/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Kyle Nitzsche <kyle.nitzsche@canonical.com>
 *              Scott Sweeny <scott.sweeny@canonical.com>
 *
 */

#include "query.h"
#include "aggchildscope.h"
#include "i18n.h"

#include <unity/UnityExceptions.h>

#include <QFile>
#include <QJsonParseError>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonArray>

using namespace unity::scopes;

QString qstr(std::string str)
{
    return QString::fromStdString(str);
}

std::string sstr(QString str)
{
    return str.toStdString();
}

/*
 * makes the AggChildScope instances using values and options
 * declared in json
 */

void Query::make_child_scopes(std::string query_string)
{

    getChildScopes();
/*
    for (std::map<std::string,std::string>::iterator it = common_templates.begin(); it != common_templates.end(); ++it)
    {
        qDebug() << QString("==== id: %1").arg(qstr(it->first));
        qDebug() << QString("==== template: %1").arg(qstr(it->second));

    }
*/
    for (std::shared_ptr<child_scope> child : child_scopes)
    {

        AggChildScope ch(child->id, query_string, child->link_to_child);

        if (child->category_title_ != "")
           ch.set_category_title(_(child->category_title_.c_str()));
        else
           ch.set_category_title(child->category_title);

        if (child->child_department != "")
           ch.set_child_department(child->child_department);

        if (child->child_category != "")
           ch.set_child_category(child->child_category);

        if (child->category_max_results != -1)
           ch.set_category_max_results(child->category_max_results);

        ch.set_cardinality(child->cardinality); // default defined in .h

        ch.set_surface_template(child->surface_template);

        if (child->search_template != "")
           ch.set_search_template(child->search_template);
        else
           ch.set_search_template(child->surface_template);

        scopes_m[ch.id()] = std::make_shared<AggChildScope>(ch);
        scope_order.emplace_back(ch.id());

    }
    return;
}

/*
 * Get child scopes from data/child_scopes.json and populate child_scopes vector
 */
void Query::getChildScopes()
{

    QString child_scopes_fn_qs = QString("%1/child_scopes.json").arg(qstr(scope_dir_));

    QFile child_scopes_f;
    child_scopes_f.setFileName(child_scopes_fn_qs);
    child_scopes_f.open(QIODevice::ReadOnly);
    QByteArray data = child_scopes_f.readAll();

    QJsonParseError err;
    QJsonDocument doc = QJsonDocument::fromJson(data, &err);

    if (err.error != QJsonParseError::NoError)
    {
        qCritical() << "Failed to parse child_scopes.json data: " << err.errorString();
        return;
    }

    QJsonObject doc_o = doc.object();
    QJsonArray common_templates_a = doc_o["common_templates"].toArray();

    for (const auto item_ : common_templates_a)
    {
        QJsonObject template_o = item_.toObject();
        std::string id = template_o["id"].toString().toStdString();

        QJsonObject template__ = template_o["template"].toObject();
        QJsonDocument t_d(template__);
        std::string template_ = sstr(t_d.toJson());

        //qDebug() << "==== template__: " <<qstr(template_);
        common_templates[id] = template_;
        //qDebug() << "==== common-t, id key: " << qstr(common_templates.find(id)->first);
        //qDebug() << "==== common-t, id val: " << qstr(common_templates.find(id)->second);

    }
    QJsonArray child_scopes_a = doc_o["child_scopes"].toArray();

    for (const auto cs_o_ : child_scopes_a)
    {
        QJsonObject cs_o = cs_o_.toObject();
        QJsonObject scope_o = cs_o["scope"].toObject();
        child_scope cs_;
        std::shared_ptr<child_scope> cs = std::make_shared<child_scope>(cs_);

        cs->id = scope_o["id"].toString().toStdString();

        if (scope_o.contains("link_to_child"))
            if (scope_o["link_to_child"].toString() == "false")
                cs->link_to_child = false;

        if (scope_o.contains("category_title"))
            cs->category_title = scope_o["category_title"].toString().toStdString();

        if (scope_o.contains("_category_title"))
            cs->category_title_ = scope_o["_category_title"].toString().toStdString();

        if (scope_o.contains("child_department"))
            cs->child_department = scope_o["child_department"].toString().toStdString();

        if (scope_o.contains("child_category"))
            cs->child_category = scope_o["child_category"].toString().toStdString();

        if (scope_o.contains("category_max_results"))
            cs->category_max_results = (int)scope_o["category_max_results"].toInt();

        if (scope_o.contains("cardinality"))
            cs->cardinality = (int)scope_o["cardinality"].toInt();

        QJsonObject surface_template_o = scope_o["surface_template"].toObject();
        QJsonDocument st_d(surface_template_o);
        cs->surface_template = sstr(st_d.toJson());

        if (scope_o.contains("search_template"))
        {
            QJsonObject search_template_o = scope_o["search_template"].toObject();
            QJsonDocument search_d(search_template_o);
            cs->search_template = sstr(search_d.toJson());
        }
        else
            cs->search_template = cs->surface_template;

        if (scope_o.contains("swap_result_attributes"))
        {
            QJsonArray swaps_a = scope_o["swap_result_attributes"].toArray();
            for (const auto swap_o_ : swaps_a)
            {
                QJsonObject swap_o = swap_o_.toObject();
                std::pair<std::string,std::string> swap;
                swap.first = swap_o["incoming_result_key"].toString().toStdString();
                swap.second = swap_o["new_result_key"].toString().toStdString();
                std::shared_ptr<std::pair<std::string,std::string>> swap_ = std::make_shared<std::pair<std::string,std::string>>(swap);
                cs->swap_result_attributes.emplace_back(swap_);
            }

        }

        if (scope_o.contains("result_category_id_to_common_template"))
        {
            QJsonArray items_a = scope_o["result_category_id_to_common_template"].toArray();
            for (const auto item_o_ : items_a)
            {
                QJsonObject item_o = item_o_.toObject();
                std::string res_cat_id = item_o["result_category_id"].toString().toStdString();
                std::string common_template = item_o["common_template"].toString().toStdString();
                cs->result_category_id_to_common_template[res_cat_id] = common_template;
                //qDebug() << "==== res_cat_id: " << qstr(res_cat_id);
                //qDebug() << "==== res_cat_id common template: " << qstr(common_template);
            }
        }
        child_scopes.emplace_back(cs);
        child_scopes_m[cs->id] = cs;
    }
}





/*
 * Loops through AggChildScopes and, for those that are enabled in settings,
 * complete initializations
 */
void Query::handle_enabled_child_scopes(bool empty_search, us::SearchReplyProxy const& upstream_reply)
{

    // Get settings as 'config'
    const us::VariantMap config = settings();
    if (config.empty())
        qWarning() << "CONFIG EMPTY!";

    // Cycle through child scopes (in scope_order) and, if enabled in settings, complete set up
    try
    {
        for (const std::string &id : scope_order)
        {
            // get a pointer to the scope
            std::shared_ptr<AggChildScope> scope = scopes_m[id];

             // if not enabled in settings, skip to next
            if (! config.at(id).get_bool())
                continue;

            // set the AggChildScope proxy from the registry
            scope->set_proxy(registry_);

            // register the categories (either for surface or for seach: they can be same)
            // if the scope uses a linkToChild category title, register with cannedQuery, else don't
            if (scope->category_link_to_child())
            {
                if (empty_search)
                {
                    scope->set_category(upstream_reply->register_category(id, scope->category_title(), "",
                                                      *(scope->canned_query()),
                                                      CategoryRenderer(scope->surface_template())));
                }
                else
                    scope->set_category(upstream_reply->register_category(id, scope->category_title(), "",
                                                  CategoryRenderer(scope->surface_template())));
            }
            else
            {
                if (empty_search)
                {
                    scope->set_category(upstream_reply->register_category(id, scope->category_title(), "",
                                                      CategoryRenderer(scope->search_template())));
                }
                else
                {
                    scope->set_category(upstream_reply->register_category(id, scope->category_title(), "",
                                                  CategoryRenderer(scope->search_template())));
                }
            }
        }
    }
    catch (unity::Exception const& e)
    {
        qWarning() << "NotFoundException: " << e.what();
    }
}

bool Query::check_hints()
{
    bool hidding_hints = false;
    if (query().query_string() == HINTS_OOBE_HIDE) {
        qDebug() << "==== HIDING HINTS";
        hidding_hints = true;
        hide_hints();
    }
    return hidding_hints;
}

bool  Query::handle_hints(us::SearchReplyProxy const& upstream_reply_)
{
    bool run_hints_only = false;
    us::ScopeProxy hints_scope_;
    auto scopes = registry_->list();
    us::MetadataMap::const_iterator scope;
    scope = scopes.find(HINTS_ID);
    if (scope != scopes.end()) {
        auto meta = registry_->get_metadata(HINTS_ID);
        hints_scope_ = meta.proxy();
    }

    if (is_hints_showing() && hints_scope_)
    {
        qDebug() << "===== Displaying Hints";
        run_hints_only = true;
        Category::SCPtr hint_cat;
        SearchMetadata metadata(search_metadata());

        auto reply = std::make_shared<ResultForwarder>(upstream_reply_,
            [this, &hint_cat](CategorisedResult& res) -> bool {
                                    return true;
        });

        subsearch(hints_scope_, "", HINTS_OOBE, FilterState(),
                  metadata, reply);
    }
    return run_hints_only;
}

bool Query::is_hints_showing(){
    auto filepath = QString::fromStdString(cache_dir_) +"/" + hints_is_hidden;
    qDebug() << QString("If this exists, we hide hints: %1. exists: %2: ").arg(filepath)<< QFile::exists(filepath);
    return !QFile::exists(filepath);
}

void Query::hide_hints() {
    auto filepath= QString::fromStdString(cache_dir_) +"/" + hints_is_hidden;
    QFile file(filepath);
    file.open(QIODevice::WriteOnly);
    QString msg = QString("Timestamp for %1").arg(qstr(THIS_SCOPE_ID));
    file.write(sstr(msg).c_str(), sstr(msg).size());
    file.close();
}



